@@ -27,24 +27,31 @@ import warnings
 import weakref
 
 try:
-    from typing import Any, Callable, Dict, Iterator, List, Set, Tuple
+    from typing import (Any, Callable, Dict, Iterator, List, Optional,
+                        Set, Tuple, Union, cast, KeysView)
     Any  # pyflakes
     Callable  # pyflakes
     Dict  # pyflakes
     Iterator  # pyflakes
+    KeysView  # pyflakes
     List  # pyflakes
+    Optional  # pyflakes
     Set  # pyflakes
     Tuple  # pyflakes
+    Union  # pyflakes
 except ImportError:
+    def cast(typ, obj):  # type: ignore
+        return obj
     pass
 
 import apt_pkg
-from apt.package import Package
+from apt.package import Package, Version
 import apt.progress.text
 from apt.progress.base import AcquireProgress, InstallProgress, OpProgress
 OpProgress  # pyflakes
 InstallProgress  # pyflakes
 AcquireProgress  # pyflakes
+Version  # pyflakes
 
 
 class FetchCancelledException(IOError):
@@ -91,15 +98,16 @@ class Cache(object):
 
     def __init__(self, progress=None, rootdir=None, memonly=False):
         # type: (OpProgress, str, bool) -> None
-        self._cache = None  # type: apt_pkg.Cache
-        self._depcache = None  # type: apt_pkg.DepCache
-        self._records = None  # type: apt_pkg.PackageRecords
-        self._list = None  # type: apt_pkg.SourceList
-        self._callbacks = {}  # type: Dict[str, List[Callable[..., None]]]
-        self._callbacks2 = {}  # type: Dict[str, List[Tuple[Callable[..., None], List[Any], Dict[Any,Any]]]] # nopep8
-        self._weakref = weakref.WeakValueDictionary()  # type: ignore
+        self._cache = cast(apt_pkg.Cache, None)  # type: apt_pkg.Cache
+        self._depcache = cast(apt_pkg.DepCache, None)  # type: apt_pkg.DepCache
+        self._records = cast(apt_pkg.PackageRecords, None)  # type: apt_pkg.PackageRecords # nopep8
+        self._list = cast(apt_pkg.SourceList, None)  # type: apt_pkg.SourceList
+        self._callbacks = {}  # type: Dict[str, List[Union[Callable[..., None],str]]] # nopep8
+        self._callbacks2 = {}  # type: Dict[str, List[Tuple[Callable[..., Any], Tuple[Any, ...], Dict[Any,Any]]]] # nopep8
+        self._weakref = weakref.WeakValueDictionary()  # type: weakref.WeakValueDictionary[str, apt.Package] # nopep8
+        self._weakversions = weakref.WeakSet()  # type: weakref.WeakSet[Version] # nopep8
         self._changes_count = -1
-        self._sorted_set = None  # type: List[str]
+        self._sorted_set = None  # type: Optional[List[str]]
 
         self.connect("cache_post_open", "_inc_changes_count")
         self.connect("cache_post_change", "_inc_changes_count")
@@ -129,6 +137,11 @@ class Cache(object):
             apt_pkg.init_system()
         self.open(progress)
 
+    def fix_broken(self):
+        # type: () -> None
+        """Fix broken packages."""
+        self._depcache.fix_broken()
+
     def _inc_changes_count(self):
         # type: () -> None
         """Increase the number of changes"""
@@ -164,7 +177,7 @@ class Cache(object):
                 if callback == '_inc_changes_count':
                     self._inc_changes_count()
                 else:
-                    callback()
+                    callback()  # type: ignore
 
         if name in self._callbacks2:
             for callback, args, kwds in self._callbacks2[name]:
@@ -188,19 +201,51 @@ class Cache(object):
         self._list = apt_pkg.SourceList()
         self._list.read_main_list()
         self._sorted_set = None
-        self._weakref.clear()
+        self.__remap()
 
         self._have_multi_arch = len(apt_pkg.get_architectures()) > 1
 
         progress.done()
         self._run_callbacks("cache_post_open")
 
+    def __remap(self):
+        # type: () -> None
+        """Called after cache reopen() to relocate to new cache.
+
+        Relocate objects like packages and versions from the old
+        underlying cache to the new one.
+        """
+        for key in list(self._weakref.keys()):
+            try:
+                pkg = self._weakref[key]
+            except KeyError:
+                continue
+
+            try:
+                pkg._pkg = self._cache[pkg._pkg.name, pkg._pkg.architecture]
+            except LookupError:
+                del self._weakref[key]
+
+        for ver in list(self._weakversions):
+            # Package has been reseated above, reseat version
+            for v in ver.package._pkg.version_list:
+                # Requirements as in debListParser::SameVersion
+                if (v.hash == ver._cand.hash and
+                    (v.size == 0 or ver._cand.size == 0 or
+                     v.size == ver._cand.size) and
+                    v.multi_arch == ver._cand.multi_arch and
+                    v.ver_str == ver._cand.ver_str):
+                    ver._cand = v
+                    break
+            else:
+                self._weakversions.remove(ver)
+
     def close(self):
         # type: () -> None
         """ Close the package cache """
         # explicitely free the FDs that _records has open
         del self._records
-        self._records = None
+        self._records = cast(apt_pkg.PackageRecords, None)
 
     def __enter__(self):
         # type: () -> Cache
@@ -208,6 +253,7 @@ class Cache(object):
         return self
 
     def __exit__(self, exc_type, exc_value, traceback):
+        # type: (object, object, object) -> None
         """ Exit the with statement """
         self.close()
 
@@ -215,22 +261,18 @@ class Cache(object):
         # type: (object) -> Package
         """ look like a dictionary (get key) """
         try:
-            return self._weakref[key]
-        except KeyError:
             key = str(key)
-            try:
-                rawpkg = self._cache[key]
-            except KeyError:
-                raise KeyError('The cache has no package named %r' % key)
+            rawpkg = self._cache[key]
+        except KeyError:
+            raise KeyError('The cache has no package named %r' % key)
 
-            # It might be excluded due to not having a version or something
-            if not self.__is_real_pkg(rawpkg):
-                raise KeyError('The cache has no package named %r' % key)
+        # It might be excluded due to not having a version or something
+        if not self.__is_real_pkg(rawpkg):
+            raise KeyError('The cache has no package named %r' % key)
 
-            pkg = self._rawpkg_to_pkg(rawpkg)
-            self._weakref[key] = pkg
+        pkg = self._rawpkg_to_pkg(rawpkg)
 
-            return pkg
+        return pkg
 
     def get(self, key, default=None):
         # type: (object, object) -> Any
@@ -244,17 +286,14 @@ class Cache(object):
             return default
 
     def _rawpkg_to_pkg(self, rawpkg):
+        # type: (apt_pkg.Package) -> Package
         """Returns the apt.Package object for an apt_pkg.Package object.
 
         .. versionadded:: 1.0.0
         """
-        fullname = rawpkg.get_fullname(pretty=False)
-        try:
-            pkg = self._weakref[fullname]
-        except KeyError:
-            pkg = Package(self, rawpkg)
-            self._weakref[fullname] = pkg
-        return pkg
+        fullname = rawpkg.get_fullname(pretty=True)
+
+        return self._weakref.setdefault(fullname, Package(self, rawpkg))
 
     def __iter__(self):
         # type: () -> Iterator[Package]
@@ -264,9 +303,11 @@ class Cache(object):
         # is disastrous if we use compressed package indexes, and slower than
         # necessary for uncompressed indexes.
         for pkgname in self.keys():
-            yield self[pkgname]
+            pkg = Package(self, self._cache[pkgname])
+            yield self._weakref.setdefault(pkgname, pkg)
 
     def __is_real_pkg(self, rawpkg):
+        # type: (apt_pkg.Package) -> bool
         """Check if the apt_pkg.Package provided is a real package."""
         return rawpkg.has_versions
 
@@ -277,7 +318,7 @@ class Cache(object):
     def __contains__(self, key):
         # type: (object) -> bool
         try:
-            return self.__is_real_pkg(self._cache[key])
+            return self.__is_real_pkg(self._cache[str(key)])
         except KeyError:
             return False
 
@@ -286,7 +327,7 @@ class Cache(object):
         return len(self.keys())
 
     def keys(self):
-        # FIXME: type: () -> List[str] - does not work
+        # type: () -> List[str]
         if self._sorted_set is None:
             self._sorted_set = sorted(p.get_fullname(pretty=True)
                                       for p in self._cache.packages
@@ -378,21 +419,22 @@ class Cache(object):
             raise CacheClosedException(
                 "Cache object used after close() called")
 
-        # get lock
-        lockfile = apt_pkg.config.find_dir("Dir::Cache::Archives") + "lock"
-        lock = apt_pkg.get_lock(lockfile)
-        if lock < 0:
-            raise LockFailedException("Failed to lock %s" % lockfile)
+        # this may as well throw a SystemError exception
+        if not pm.get_archives(fetcher, self._list, self._records):
+            return False
+        # now run the fetcher, throw exception if something fails to be
+        # fetched
+        return self._run_fetcher(fetcher)
 
+    def _get_archive_lock(self, fetcher):
+        # type: (apt_pkg.Acquire) -> None
+        # get lock
+        archive_dir = apt_pkg.config.find_dir("Dir::Cache::Archives")
         try:
-            # this may as well throw a SystemError exception
-            if not pm.get_archives(fetcher, self._list, self._records):
-                return False
-            # now run the fetcher, throw exception if something fails to be
-            # fetched
-            return self._run_fetcher(fetcher)
-        finally:
-            os.close(lock)
+            fetcher.get_lock(archive_dir)
+        except apt_pkg.Error as e:
+            raise LockFailedException(("Failed to lock archive directory %s: "
+                                       " %s") % (archive_dir, e))
 
     def fetch_archives(self, progress=None, fetcher=None):
         # type: (AcquireProgress, apt_pkg.Acquire) -> int
@@ -415,6 +457,8 @@ class Cache(object):
         if fetcher is None:
             fetcher = apt_pkg.Acquire(progress)
 
+        self._get_archive_lock(fetcher)
+
         return self._fetch_archives(fetcher,
                                     apt_pkg.PackageManager(self._depcache))
 
@@ -463,7 +507,7 @@ class Cache(object):
 
     def update(self, fetch_progress=None, pulse_interval=0,
                raise_on_error=True, sources_list=None):
-        # FIXME: type: (AcquireProgress, int, bool, str) -> int
+        # type: (AcquireProgress, int, bool, str) -> int
         """Run the equivalent of apt-get update.
 
         You probably want to call open() afterwards, in order to utilise the
@@ -522,13 +566,26 @@ class Cache(object):
 
         The second parameter *install_progress* refers to an InstallProgress()
         object of the module apt.progress.
+
+        This releases a system lock in newer versions, if there is any,
+        and reestablishes it afterwards.
         """
         # compat with older API
         try:
             install_progress.startUpdate()  # type: ignore
         except AttributeError:
             install_progress.start_update()
-        res = install_progress.run(pm)
+
+        did_unlock = apt_pkg.pkgsystem_is_locked()
+        if did_unlock:
+            apt_pkg.pkgsystem_unlock_inner()
+
+        try:
+            res = install_progress.run(pm)
+        finally:
+            if did_unlock:
+                apt_pkg.pkgsystem_lock_inner()
+
         try:
             install_progress.finishUpdate()  # type: ignore
         except AttributeError:
@@ -558,25 +615,30 @@ class Cache(object):
         if install_progress is None:
             install_progress = apt.progress.base.InstallProgress()
 
-        pm = apt_pkg.PackageManager(self._depcache)
-        fetcher = apt_pkg.Acquire(fetch_progress)
-        while True:
-            # fetch archives first
-            res = self._fetch_archives(fetcher, pm)
+        assert install_progress is not None
 
-            # then install
-            res = self.install_archives(pm, install_progress)
-            if res == pm.RESULT_COMPLETED:
-                break
-            elif res == pm.RESULT_FAILED:
-                raise SystemError("installArchives() failed")
-            elif res == pm.RESULT_INCOMPLETE:
-                pass
-            else:
-                raise SystemError("internal-error: unknown result code "
-                                  "from InstallArchives: %s" % res)
-            # reload the fetcher for media swaping
-            fetcher.shutdown()
+        with apt_pkg.SystemLock():
+            pm = apt_pkg.PackageManager(self._depcache)
+            fetcher = apt_pkg.Acquire(fetch_progress)
+            self._get_archive_lock(fetcher)
+
+            while True:
+                # fetch archives first
+                res = self._fetch_archives(fetcher, pm)
+
+                # then install
+                res = self.install_archives(pm, install_progress)
+                if res == pm.RESULT_COMPLETED:
+                    break
+                elif res == pm.RESULT_FAILED:
+                    raise SystemError("installArchives() failed")
+                elif res == pm.RESULT_INCOMPLETE:
+                    pass
+                else:
+                    raise SystemError("internal-error: unknown result code "
+                                      "from InstallArchives: %s" % res)
+                # reload the fetcher for media swaping
+                fetcher.shutdown()
         return (res == pm.RESULT_COMPLETED)
 
     def clear(self):
@@ -598,6 +660,7 @@ class Cache(object):
         self._run_callbacks("cache_pre_change")
 
     def connect(self, name, callback):
+        # type: (str, Union[Callable[..., None],str]) -> None
         """Connect to a signal.
 
         .. deprecated:: 1.0
@@ -613,6 +676,7 @@ class Cache(object):
         self._callbacks[name].append(callback)
 
     def connect2(self, name, callback, *args, **kwds):
+        # type: (str, Callable[..., Any], object, object) -> None
         """Connect to a signal.
 
         The callback will be passed the cache as an argument, and
@@ -841,7 +905,7 @@ class FilteredCache(object):
             yield self.cache[pkgname]
 
     def keys(self):
-        # FIXME: type: () -> List[str] - does not work
+        # type: () -> KeysView[str]
         return self._helper._filtered.keys()
 
     def has_key(self, key):
@@ -868,19 +932,23 @@ class FilteredCache(object):
         self._helper.filter_cache_post_change(self.cache)
 
     def __getattr__(self, key):
+        # type: (str) -> Any
         """we try to look exactly like a real cache."""
         return getattr(self.cache, key)
 
 
 def cache_pre_changed(cache):
+    # type: (Cache) -> None
     print("cache pre changed")
 
 
 def cache_post_changed(cache):
+    # type: (Cache) -> None
     print("cache post changed")
 
 
 def _test():
+    # type: () -> None
     """Internal test code."""
     print("Cache self test")
     apt_pkg.init()
@@ -919,7 +987,7 @@ def _test():
     filtered.set_filter(MarkedChangesFilter())
     print(len(filtered))
     for pkgname in filtered.keys():
-        assert pkgname == filtered[pkg].name
+        assert pkgname == filtered[pkgname].name
 
     print(len(filtered))
 
